import pydantic.version

if int(pydantic.version.VERSION[0]) <= 1:  # pragma: no cover
    raise ImportError("pydantic_compat._v2 only supports pydantic v2.x")

from .decorators import root_validator as root_validator
from .decorators import validator as validator
from .mixin import PydanticCompatMixin as PydanticCompatMixin



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, cast

from pydantic import BaseModel
from pydantic._internal import _model_construction

from pydantic_compat._shared import V2_RENAMED_CONFIG_KEYS, check_mixin_order

if TYPE_CHECKING:
    from pydantic import ConfigDict
    from typing_extensions import Protocol

    # fmt:off
    class Model(Protocol):
        def model_dump(self, *args: Any, **kwargs: Any) ->  dict[str, Any]: ...
        def model_dump_json(self, *args: Any, **kwargs: Any) -> str: ...
        def model_copy(self, *args: Any, **kwargs: Any) -> Model: ...
        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> dict[str, Any]: ...
        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Model: ...
        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Model: ...
        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> type[Model]: ...
        @classmethod
        def model_rebuild(cls, *args: Any, **kwargs: Any) -> bool | None: ...

        model_fields: ClassVar[dict]
        model_fields_set: ClassVar[set[str]]
        model_config: ClassVar[ConfigDict]
    # fmt:on


def _convert_config(config: type) -> ConfigDict:
    config_dict = {k: getattr(config, k) for k in dir(config) if not k.startswith("__")}

    deprecated_renamed_keys = V2_RENAMED_CONFIG_KEYS.keys() & config_dict.keys()
    for k in sorted(deprecated_renamed_keys):
        config_dict[V2_RENAMED_CONFIG_KEYS[k]] = config_dict.pop(k)

    # leave these here for now to warn about lost functionality
    # deprecated_removed_keys = V2_REMOVED_CONFIG_KEYS & config_dict.keys()
    # for k in sorted(deprecated_removed_keys):
    #     config_dict.pop(k)

    return cast("ConfigDict", config_dict)


class _MixinMeta(_model_construction.ModelMetaclass):
    def __new__(cls, name, bases, namespace, **kwargs):  # type: ignore
        if "Config" in namespace and isinstance(namespace["Config"], type):
            namespace["model_config"] = _convert_config(namespace.pop("Config"))
        return super().__new__(cls, name, bases, namespace, **kwargs)


class PydanticCompatMixin(metaclass=_MixinMeta):
    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        check_mixin_order(cls, PydanticCompatMixin, BaseModel)
        # the deprecation warning is on the metaclass
        type(cls).__fields__ = property(lambda cls: cls.model_fields)  # type: ignore

    def dict(self: Model, *args: Any, **kwargs: Any) -> Any:
        return self.model_dump(*args, **kwargs)

    def json(self: Model, *args: Any, **kwargs: Any) -> Any:
        return self.model_dump_json(*args, **kwargs)

    def copy(self: Model, *args: Any, **kwargs: Any) -> Any:
        return self.model_copy(*args, **kwargs)

    @classmethod
    def schema(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.model_json_schema(*args, **kwargs)

    @classmethod
    def validate(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.model_validate(*args, **kwargs)

    @classmethod
    def construct(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.model_construct(*args, **kwargs)

    @classmethod
    def parse_obj(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.model_validate(*args, **kwargs)

    @classmethod
    def parse_raw(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.model_validate_json(*args, **kwargs)

    # this is needed in addition to the metaclass patch in __init_subclass__
    @property
    def __fields__(self: Model) -> Dict[str, Any]:  # noqa: UP006
        return self.model_fields

    @property
    def __fields_set__(self: Model) -> set[str]:
        return self.model_fields_set

    @classmethod
    def update_forward_refs(
        cls: type[Model],
        force: bool = False,
        raise_errors: bool = True,
        **localns: Any,
    ) -> None:
        cls.model_rebuild(
            forc=force, raise_errors=raise_errors, _types_namespace=localns
        )

    @classmethod
    def model_rebuild(
        cls: type[Model], force: bool = False, raise_errors: bool = True, **kwargs: Any
    ) -> bool | None:
        return super().model_rebuild(
            force=force, raise_errors=raise_errors, _types_namespace=kwargs
        )



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable

import pydantic

if TYPE_CHECKING:
    from typing import Literal


# V2 signature
def field_validator(
    _field: str,
    *fields: str,
    mode: Literal["before", "after", "wrap", "plain"] = "after",
    check_fields: bool | None = None,
) -> Callable:
    """Adaptor from v2.field_validator -> v1.validator."""
    # V1 signature
    # def validator(
    #     *fields: str,
    #     pre: bool = False,
    #     each_item: bool = False,
    #     always: bool = False,
    #     check_fields: bool = True,
    #     whole: Optional[bool] = None,
    #     allow_reuse: bool = False,
    # ) -> Callable[[AnyCallable], 'AnyClassMethod']:
    #     ...
    return pydantic.validator(
        _field,
        *fields,
        pre=(mode in ("before")),
        always=True,  # should it be?
        check_fields=bool(check_fields),
        allow_reuse=True,
    )


# V2 signature
def model_validator(*, mode: Literal["wrap", "before", "after"]) -> Any:
    """Adaptor from v2.model_validator -> v1.root_validator."""

    # V1 signature
    # def root_validator(
    #     _func: Optional[AnyCallable] = None,
    #     *,
    #     pre: bool = False,
    #     allow_reuse: bool = False,
    #     skip_on_failure: bool = False,
    # ) -> Union["AnyClassMethod", Callable[[AnyCallable], "AnyClassMethod"]]:
    #     ...
    return root_validator(
        pre=mode == "before", allow_reuse=True, construct_object=mode == "after"
    )


def root_validator(
    _func: Callable | None = None,
    *,
    pre: bool = False,
    allow_reuse: bool = False,
    skip_on_failure: bool = False,
    construct_object: bool = False,
) -> Any:
    def _inner(_func: Callable) -> Any:
        func = _func
        if construct_object and not pre:
            if isinstance(_func, classmethod):
                _func = _func.__func__

            @wraps(_func)
            def func(cls: type[pydantic.BaseModel], *args: Any, **kwargs: Any) -> Any:
                arg0, *rest = args
                # cast dict to model to match the v2 model_validator signature
                # using construct because it should already be valid
                new_args = (cls.construct(**arg0), *rest)
                result: pydantic.BaseModel = _func(cls, *new_args, **kwargs)
                # cast back to dict of field -> value
                return {k: getattr(result, k) for k in result.__fields__}

        deco = pydantic.root_validator(  # type: ignore [call-overload]
            pre=pre, allow_reuse=allow_reuse, skip_on_failure=skip_on_failure
        )
        return deco(func)

    return _inner(_func) if _func else _inner



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, ClassVar, Iterator, Mapping

from pydantic import main

from pydantic_compat._shared import V2_RENAMED_CONFIG_KEYS, check_mixin_order

if TYPE_CHECKING:
    from typing import Dict

    from pydantic.fields import ModelField  # type: ignore
    from typing_extensions import Protocol

    # fmt:off
    class Model(Protocol):
        def dict(self, *args: Any, **kwargs: Any) ->  Dict[str, Any]: ...  # noqa: UP006
        def json(self, *args: Any, **kwargs: Any) -> str: ...
        def copy(self, *args: Any, **kwargs: Any) -> Model: ...
        @classmethod
        def schema(cls, *args: Any, **kwargs: Any) -> Dict[str, Any]: ...  # noqa: UP006
        @classmethod
        def validate(cls, *args: Any, **kwargs: Any) -> Model: ...
        @classmethod
        def construct(cls, *args: Any, **kwargs: Any) -> Model: ...
        @classmethod
        def parse_raw(cls, *args: Any, **kwargs: Any) -> type[Model]: ...
        @classmethod
        def update_forward_refs(cls, *args: Any, **kwargs: Any) -> None: ...

        __fields__: ClassVar[Dict]  # noqa: UP006
        __fields_set__: set[str]
        __config__: ClassVar[type]
    # fmt:on

if sys.version_info < (3, 9):

    def _get_fields(obj) -> dict[str, Any]:
        return obj.__fields__

    main.ModelMetaclass.model_fields = property(_get_fields)

REVERSE_CONFIG_NAME_MAP = {v: k for k, v in V2_RENAMED_CONFIG_KEYS.items()}


def _convert_config(config_dict: dict) -> type:
    deprecated_renamed_keys = REVERSE_CONFIG_NAME_MAP.keys() & config_dict.keys()
    for k in sorted(deprecated_renamed_keys):
        config_dict[REVERSE_CONFIG_NAME_MAP[k]] = config_dict.pop(k)

    return type("Config", (), config_dict)


class _MixinMeta(main.ModelMetaclass):
    def __new__(cls, name, bases, namespace: dict, **kwargs):  # type: ignore
        if "model_config" in namespace and isinstance(namespace["model_config"], dict):
            namespace["Config"] = _convert_config(namespace.pop("model_config"))

        return super().__new__(cls, name, bases, namespace, **kwargs)


class PydanticCompatMixin(metaclass=_MixinMeta):
    @classmethod
    def __try_update_forward_refs__(cls, **localns: Any) -> None:
        sup = super()
        if hasattr(sup, "__try_update_forward_refs__"):
            sup.__try_update_forward_refs__(**localns)

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        check_mixin_order(cls, PydanticCompatMixin, main.BaseModel)

    def model_dump(self: Model, *args: Any, **kwargs: Any) -> Any:
        return self.dict(*args, **kwargs)

    def model_dump_json(self: Model, *args: Any, **kwargs: Any) -> Any:
        return self.json(*args, **kwargs)

    def model_copy(self: Model, *args: Any, **kwargs: Any) -> Any:
        return self.copy(*args, **kwargs)

    @classmethod
    def model_json_schema(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.schema(*args, **kwargs)

    @classmethod
    def model_validate(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.validate(*args, **kwargs)

    @classmethod
    def model_construct(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.construct(*args, **kwargs)

    @classmethod
    def model_validate_json(cls: type[Model], *args: Any, **kwargs: Any) -> Any:
        return cls.parse_raw(*args, **kwargs)

    @classmethod
    def model_rebuild(cls: type[Model], force: bool = True, **kwargs: Any) -> None:
        return cls.update_forward_refs(**kwargs)

    if sys.version_info < (3, 9):
        # differences in the behavior of patching class properties in python<3.9
        @property
        def model_fields(cls: type[Model]) -> Mapping[str, Any]:
            return FieldInfoMap(cls.__fields__)

    else:

        @classmethod  # type: ignore [misc]
        @property
        def model_fields(cls: type[Model]) -> Mapping[str, Any]:
            return FieldInfoMap(cls.__fields__)

    @property
    def model_fields_set(self: Model) -> set[str]:
        return self.__fields_set__

    @classmethod  # type: ignore [misc]
    @property
    def model_config(cls: type[Model]) -> Mapping[str, Any]:
        return DictLike(cls.__config__)


class FieldInfoLike:
    """Wrapper to convera pydantic v1 ModelField to v2 FieldInfo."""

    def __init__(self, model_field: ModelField) -> None:
        self._model_field = model_field

    @property
    def annotation(self) -> Any:
        return self._model_field.outer_type_

    @property
    def frozen(self) -> bool:
        return not self._model_field.field_info.allow_mutation

    def __getattr__(self, key: str) -> Any:
        return getattr(self._model_field, key)


class FieldInfoMap(Mapping[str, FieldInfoLike]):
    """Adaptor between v1 __fields__ and v2 model_field."""

    def __init__(self, fields: dict[str, ModelField]) -> None:
        self._fields = fields

    def get(self, key: str, default: Any = None) -> Any:
        return self[key] if key in self._fields else default

    def __getitem__(self, key: str) -> FieldInfoLike:
        return FieldInfoLike(self._fields[key])

    def __setitem__(self, key: str, value: Any) -> None:
        self._fields[key] = value

    def __iter__(self) -> Iterator[str]:
        yield from self._fields

    def __len__(self) -> int:
        return len(self._fields)


class DictLike(Mapping[str, Any]):
    """Provide dict-like interface to an object."""

    def __init__(self, obj: Any) -> None:
        self._obj = obj

    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self._obj, key, default)

    def __getitem__(self, key: str) -> Any:
        return getattr(self._obj, key)

    def __setitem__(self, key: str, value: Any) -> None:
        setattr(self._obj, key, value)

    def __iter__(self) -> Iterator[str]:
        yield from self._obj.__dict__

    def __len__(self) -> int:
        return len(self._obj.__dict__)
